# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators
import lale.type_checking
from lale.type_checking import is_schema


def _columns_schema_to_list(X, schema):
    s_all = lale.datasets.data_schemas.to_schema(X)
    s_row = s_all["items"]
    n_columns = s_row["minItems"]
    assert n_columns == s_row["maxItems"]
    s_cols = s_row["items"]
    if isinstance(s_cols, dict):
        if lale.type_checking.is_subschema(s_cols, schema):
            result = [*range(n_columns)]
        else:
            result = []
    else:
        assert isinstance(s_cols, list)
        result = [
            i
            for i in range(n_columns)
            if lale.type_checking.is_subschema(s_cols[i], schema)
        ]
    return result


def _columns_to_list(columns, kind, X):
    if columns is None:
        if kind == "passthrough":
            result = [*range(X.shape[1])]
        else:
            result = []
    elif isinstance(columns, list):
        result = columns
    elif callable(columns):
        result = columns(X)
    elif is_schema(columns):
        result = _columns_schema_to_list(X, columns)
    else:
        raise TypeError(f"type {type(columns)}, columns {columns}")
    if len(result) > 0 and isinstance(result[0], str):
        name2idx = {name: idx for idx, name in enumerate(X.columns)}
        result = [name2idx[name] for name in result]
    return result


class _ProjectImpl:
    def __init__(self, columns=None, drop_columns=None):
        self._hyperparams = {"columns": columns, "drop_columns": drop_columns}

    def fit(self, X, y=None):
        keep_cols = self._hyperparams["columns"]
        keep_list = _columns_to_list(keep_cols, "passthrough", X)
        drop_cols = self._hyperparams["drop_columns"]
        drop_list = _columns_to_list(drop_cols, "drop", X)
        self._fit_columns = [c for c in keep_list if c not in drop_list]
        return self

    def transform(self, X):
        if isinstance(X, pd.DataFrame):
            if len(self._fit_columns) == 0 or isinstance(self._fit_columns[0], int):
                result = X.iloc[:, self._fit_columns]
            else:
                result = X[self._fit_columns]
        elif isinstance(X, np.ndarray):
            result = X[:, self._fit_columns]
        else:
            raise TypeError(f"type {type(X)}")
        s_X = lale.datasets.data_schemas.to_schema(X)
        s_result = self.transform_schema(s_X)
        result = lale.datasets.data_schemas.add_schema(result, s_result)
        result = lale.datasets.data_schemas.add_table_name(
            result, lale.datasets.data_schemas.get_table_name(X)
        )
        return result

    def transform_schema(self, s_X):
        """Used internally by Lale for type-checking downstream operators."""
        if is_schema(s_X):
            if hasattr(self, "_fit_columns"):
                return self._transform_schema_fit_columns(s_X)
            keep_cols = self._hyperparams["columns"]
            drop_cols = self._hyperparams["drop_columns"]
            if (keep_cols is None or is_schema(keep_cols)) and (
                drop_cols is None or is_schema(drop_cols)
            ):
                return self._transform_schema_schema(s_X, keep_cols, drop_cols)
            return s_X
        else:
            X = lale.datasets.data_schemas.add_schema(s_X)
            assert X is not None
            self.fit(X)
            return self._transform_schema_fit_columns(X.json_schema)

    def _transform_schema_fit_columns(self, s_X):
        s_X = lale.datasets.data_schemas.to_schema(s_X)
        s_row = s_X["items"]
        s_cols = s_row["items"]
        n_columns = len(self._fit_columns)
        if isinstance(s_cols, dict):
            s_cols_result = s_cols
        else:
            name2i = {s_cols[i]["description"]: i for i in range(len(s_cols))}
            keep_cols_i = [
                name2i[col] if isinstance(col, str) else col
                for col in self._fit_columns
            ]
            s_cols_result = [s_cols[i] for i in keep_cols_i]
        s_result = {
            **s_X,
            "items": {
                **s_row,
                "minItems": n_columns,
                "maxItems": n_columns,
                "items": s_cols_result,
            },
        }
        return s_result

    def _transform_schema_schema(self, s_X, s_keep, s_drop):
        def is_keeper(column_schema):
            if s_keep is not None:
                if not lale.type_checking.is_subschema(column_schema, s_keep):
                    return False
            if s_drop is not None:
                if lale.type_checking.is_subschema(column_schema, s_drop):
                    return False
            return True

        s_X = lale.datasets.data_schemas.to_schema(s_X)
        s_row = s_X["items"]
        s_cols = s_row["items"]
        if isinstance(s_cols, dict):
            if is_keeper(s_cols):
                s_row_result = s_row
            else:
                s_row_result = {"type": "array", "minItems": 0, "maxItems": 0}
        else:
            assert isinstance(s_cols, list)
            s_cols_result = [s for s in s_cols if is_keeper(s)]
            n_columns = len(s_cols_result)
            s_row_result = {
                "type": "array",
                "minItems": n_columns,
                "maxItems": n_columns,
                "items": s_cols_result,
            }
        return {"type": "array", "items": s_row_result}


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their types, one at a time, omitting cross-argument constraints, if any.",
            "type": "object",
            "additionalProperties": False,
            "required": ["columns", "drop_columns"],
            "relevantToOptimizer": [],
            "properties": {
                "columns": {
                    "description": """The subset of columns to retain.

The supported column specification formats include some of the ones
from scikit-learn's ColumnTransformer_, and in addition, filtering by
using a JSON subschema_ check.

.. _ColumnTransformer: https://scikit-learn.org/stable/modules/generated/sklearn.compose.ColumnTransformer.html
.. _subschema: https://github.com/IBM/jsonsubschema""",
                    "anyOf": [
                        {
                            "enum": [None],
                            "description": "If not specified, keep all columns.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "integer"},
                            "description": "Multiple columns by index.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "string"},
                            "description": "Multiple Dataframe columns by names.",
                        },
                        {
                            "laleType": "callable",
                            "description": "Callable that is passed the input data X and can return a list of column names or indices.",
                        },
                        {
                            "type": "object",
                            "description": "Keep columns whose schema is a subschema of this JSON schema.",
                        },
                    ],
                    "default": None,
                },
                "drop_columns": {
                    "description": """The subset of columns to remove.

The `drop_columns` argument supports the same formats as `columns`.
If both are specified, keep everything from `columns` that is not
also in `drop_columns`.""",
                    "anyOf": [
                        {
                            "enum": [None],
                            "description": "If not specified, drop no further columns.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "integer"},
                            "description": "Multiple columns by index.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "string"},
                            "description": "Multiple Dataframe columns by names.",
                        },
                        {
                            "laleType": "callable",
                            "description": "Callable that is passed the input data X and can return a list of column names or indices.",
                        },
                        {
                            "type": "object",
                            "description": "Remove columns whose schema is a subschema of this JSON schema.",
                        },
                    ],
                    "default": None,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        },
        "y": {"description": "Target for supervised learning (ignored)."},
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        }
    },
}

_output_transform_schema = {
    "description": "Features; the outer array is over samples.",
    "type": "array",
    "items": {
        "type": "array",
        "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
    },
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Projection keeps a subset of the columns, like in relational algebra.

Examples
--------
>>> df = pd.DataFrame(data={'A': [1,2], 'B': ['x','y'], 'C': [3,4]})
>>> keep_numbers = Project(columns={'type': 'number'})
>>> keep_numbers.fit(df).transform(df)
NDArrayWithSchema([[1, 3],
                   [2, 4]])
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.project.html",
    "import_from": "lale.lib.lale",
    "type": "object",
    "tags": {"pre": ["categoricals"], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


Project = lale.operators.make_operator(_ProjectImpl, _combined_schemas)

lale.docstrings.set_docstrings(Project)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import urllib.request
from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OneHotEncoder

try:
    import arff
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        """Package 'arff' not found. You can install it with
    pip install 'liac-arff>=2.4.0'
or with
    pip install 'lale[full]'"""
    )

download_data_dir = os.path.join(os.path.dirname(__file__), "download_data")
experiments_dict: Dict[str, Dict[str, str]] = {}

# 1.25
experiments_dict["vehicle"] = {}
experiments_dict["vehicle"][
    "download_arff_url"
] = "https://www.openml.org/data/download/54/dataset_54_vehicle.arff"
experiments_dict["vehicle"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/54/dataset_54_vehicle.arff"
experiments_dict["vehicle"]["task_type"] = "classification"
experiments_dict["vehicle"]["target"] = "class"

# 1.3
experiments_dict["blood-transfusion-service-center"] = {}
experiments_dict["blood-transfusion-service-center"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1586225/php0iVrYT"
experiments_dict["blood-transfusion-service-center"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1586225/php0iVrYT"
experiments_dict["blood-transfusion-service-center"]["task_type"] = "classification"
experiments_dict["blood-transfusion-service-center"]["target"] = "class"

# 1.5
experiments_dict["car"] = {}
experiments_dict["car"][
    "download_arff_url"
] = "https://www.openml.org/data/download/18116966/php2jDIhh"
experiments_dict["car"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/18116966/php2jDIhh"
experiments_dict["car"]["task_type"] = "classification"
experiments_dict["car"]["target"] = "class"

# 1.6
experiments_dict["kc1"] = {}
experiments_dict["kc1"][
    "download_arff_url"
] = "https://www.openml.org/data/download/53950/kc1.arff"
experiments_dict["kc1"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/53950/kc1.arff"
experiments_dict["kc1"]["task_type"] = "classification"
experiments_dict["kc1"]["target"] = "defects"

# 2.6
experiments_dict["Australian"] = {}
experiments_dict["Australian"][
    "download_arff_url"
] = "https://www.openml.org/data/download/18151910/phpelnJ6y"
experiments_dict["Australian"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/18151910/phpelnJ6y"
experiments_dict["Australian"]["task_type"] = "classification"
experiments_dict["Australian"]["target"] = "a15"

# 3.1
experiments_dict["credit-g"] = {}
experiments_dict["credit-g"][
    "download_arff_url"
] = "https://www.openml.org/data/download/31/dataset_31_credit-g.arff"
experiments_dict["credit-g"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/31/dataset_31_credit-g.arff"
experiments_dict["credit-g"]["task_type"] = "classification"
experiments_dict["credit-g"]["target"] = "class"

# 3.4
experiments_dict["phoneme"] = {}
experiments_dict["phoneme"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1592281/php8Mz7BG"
experiments_dict["phoneme"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1592281/php8Mz7BG"
experiments_dict["phoneme"]["task_type"] = "classification"
experiments_dict["phoneme"]["target"] = "class"

# 3.6
experiments_dict["kr-vs-kp"] = {}
experiments_dict["kr-vs-kp"][
    "download_arff_url"
] = "https://www.openml.org/data/download/3/dataset_3_kr-vs-kp.arff"
experiments_dict["kr-vs-kp"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/3/dataset_3_kr-vs-kp.arff"
experiments_dict["kr-vs-kp"]["task_type"] = "classification"
experiments_dict["kr-vs-kp"]["target"] = "class"

# 4.0
experiments_dict["mfeat-factors"] = {}
experiments_dict["mfeat-factors"][
    "download_arff_url"
] = "https://www.openml.org/data/download/12/dataset_12_mfeat-factors.arff"
experiments_dict["mfeat-factors"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/12/dataset_12_mfeat-factors.arff"
experiments_dict["mfeat-factors"]["task_type"] = "classification"
experiments_dict["mfeat-factors"]["target"] = "class"

# 5.9
experiments_dict["cnae-9"] = {}
experiments_dict["cnae-9"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1586233/phpmcGu2X"
experiments_dict["cnae-9"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1586233/phpmcGu2X"
experiments_dict["cnae-9"]["task_type"] = "classification"
experiments_dict["cnae-9"]["target"] = "class"

# 8.1
experiments_dict["sylvine"] = {}
experiments_dict["sylvine"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335519/file7a97574fa9ae.arff"
experiments_dict["sylvine"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335519/file7a97574fa9ae.arff"
experiments_dict["sylvine"]["task_type"] = "classification"
experiments_dict["sylvine"]["target"] = "class"

# 17
experiments_dict["jungle_chess_2pcs_raw_endgame_complete"] = {}
experiments_dict["jungle_chess_2pcs_raw_endgame_complete"][
    "download_arff_url"
] = "https://www.openml.org/data/download/18631418/jungle_chess_2pcs_raw_endgame_complete.arff"
experiments_dict["jungle_chess_2pcs_raw_endgame_complete"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/18631418/jungle_chess_2pcs_raw_endgame_complete.arff"
experiments_dict["jungle_chess_2pcs_raw_endgame_complete"][
    "task_type"
] = "classification"
experiments_dict["jungle_chess_2pcs_raw_endgame_complete"]["target"] = "class"

# 32
experiments_dict["shuttle"] = {}
experiments_dict["shuttle"][
    "download_arff_url"
] = "https://www.openml.org/data/download/4965262/shuttle.arff"
experiments_dict["shuttle"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/4965262/shuttle.arff"
experiments_dict["shuttle"]["task_type"] = "classification"
experiments_dict["shuttle"]["target"] = "class"

# 55
experiments_dict["jasmine"] = {}
experiments_dict["jasmine"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335516/file79b563a1a18.arff"
experiments_dict["jasmine"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335516/file79b563a1a18.arff"
experiments_dict["jasmine"]["task_type"] = "classification"
experiments_dict["jasmine"]["target"] = "class"

# 118
experiments_dict["fabert"] = {}
experiments_dict["fabert"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335687/file1c555f4ca44d.arff"
experiments_dict["fabert"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335687/file1c555f4ca44d.arff"
experiments_dict["fabert"]["task_type"] = "classification"
experiments_dict["fabert"]["target"] = "class"

# 226
experiments_dict["helena"] = {}
experiments_dict["helena"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335692/file1c556677f875.arff"
experiments_dict["helena"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335692/file1c556677f875.arff"
experiments_dict["helena"]["task_type"] = "classification"
experiments_dict["helena"]["target"] = "class"

# 230
experiments_dict["bank-marketing"] = {}
experiments_dict["bank-marketing"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1586218/phpkIxskf"
experiments_dict["bank-marketing"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1586218/phpkIxskf"
experiments_dict["bank-marketing"]["task_type"] = "classification"
experiments_dict["bank-marketing"]["target"] = "class"

# 407
experiments_dict["nomao"] = {}
experiments_dict["nomao"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1592278/phpDYCOet"
experiments_dict["nomao"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1592278/phpDYCOet"
experiments_dict["nomao"]["task_type"] = "classification"
experiments_dict["nomao"]["target"] = "class"

# 425
experiments_dict["dilbert"] = {}
experiments_dict["dilbert"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335686/file1c5552c0c4b0.arff"
experiments_dict["dilbert"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335686/file1c5552c0c4b0.arff"
experiments_dict["dilbert"]["task_type"] = "classification"
experiments_dict["dilbert"]["target"] = "class"

# 442
experiments_dict["numerai28.6"] = {}
experiments_dict["numerai28.6"][
    "download_arff_url"
] = "https://www.openml.org/data/download/2160285/phpg2t68G"
experiments_dict["numerai28.6"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/2160285/phpg2t68G"
experiments_dict["numerai28.6"]["task_type"] = "classification"
experiments_dict["numerai28.6"]["target"] = "attribute_21"

# 457
experiments_dict["prnn_cushings"] = {
    "task_type": "classification",
    "target": "type",
    "download_arff_url": "https://www.openml.org/data/download/52569/prnn_cushings.arff",
    "download_csv_url": "https://www.openml.org/data/get_csv/52569/prnn_cushings.csv",
}

# 503
experiments_dict["adult"] = {}
experiments_dict["adult"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1595261/phpMawTba"
experiments_dict["adult"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1595261/phpMawTba"
experiments_dict["adult"]["task_type"] = "classification"
experiments_dict["adult"]["target"] = "class"

# 633
experiments_dict["higgs"] = {}
experiments_dict["higgs"][
    "download_arff_url"
] = "https://www.openml.org/data/download/2063675/phpZLgL9q"
experiments_dict["higgs"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/2063675/phpZLgL9q"
experiments_dict["higgs"]["task_type"] = "classification"
experiments_dict["higgs"]["target"] = "class"

# 981
experiments_dict["christine"] = {}
experiments_dict["christine"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335515/file764d5d063390.arff"
experiments_dict["christine"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335515/file764d5d063390.arff"
experiments_dict["christine"]["task_type"] = "classification"
experiments_dict["christine"]["target"] = "class"

# 1169
experiments_dict["jannis"] = {}
experiments_dict["jannis"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335691/file1c558ee247d.arff"
experiments_dict["jannis"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335691/file1c558ee247d.arff"
experiments_dict["jannis"]["task_type"] = "classification"
experiments_dict["jannis"]["target"] = "class"

# 1503
experiments_dict["connect-4"] = {}
experiments_dict["connect-4"][
    "download_arff_url"
] = "https://www.openml.org/data/download/4965243/connect-4.arff"
experiments_dict["connect-4"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/4965243/connect-4.arff"
experiments_dict["connect-4"]["task_type"] = "classification"
experiments_dict["connect-4"]["target"] = "class"

# 1580
experiments_dict["volkert"] = {}
experiments_dict["volkert"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335689/file1c556e3db171.arff"
experiments_dict["volkert"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335689/file1c556e3db171.arff"
experiments_dict["volkert"]["task_type"] = "classification"
experiments_dict["volkert"]["target"] = "class"

# 2112
experiments_dict["APSFailure"] = {}
experiments_dict["APSFailure"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335511/aps_failure.arff"
experiments_dict["APSFailure"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335511/aps_failure.arff"
experiments_dict["APSFailure"]["task_type"] = "classification"
experiments_dict["APSFailure"]["target"] = "class"

# 3700
experiments_dict["riccardo"] = {}
experiments_dict["riccardo"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335534/file7b535210a7df.arff"
experiments_dict["riccardo"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335534/file7b535210a7df.arff"
experiments_dict["riccardo"]["task_type"] = "classification"
experiments_dict["riccardo"]["target"] = "class"

# 3759
experiments_dict["guillermo"] = {}
experiments_dict["guillermo"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335532/file7b5323e77330.arff"
experiments_dict["guillermo"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335532/file7b5323e77330.arff"
experiments_dict["guillermo"]["task_type"] = "classification"
experiments_dict["guillermo"]["target"] = "class"

experiments_dict["albert"] = {}
experiments_dict["albert"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335520/file7b53746cbda2.arff"
experiments_dict["albert"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335520/file7b53746cbda2.arff"
experiments_dict["albert"]["task_type"] = "classification"
experiments_dict["albert"]["target"] = "class"

experiments_dict["robert"] = {}
experiments_dict["robert"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335688/file1c55384ec217.arff"
experiments_dict["robert"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335688/file1c55384ec217.arff"
experiments_dict["robert"]["task_type"] = "classification"
experiments_dict["robert"]["target"] = "class"

experiments_dict["covertype"] = {}
experiments_dict["covertype"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1601911/phpQOf0wY"
experiments_dict["covertype"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1601911/phpQOf0wY"
experiments_dict["covertype"]["task_type"] = "classification"
experiments_dict["covertype"]["target"] = "class"

# This dataset doesn't work with the pre-processing pipeline coded below, as the SimpleImputer drops some columns
# which have all missing values. There is no easy way to pass this info to the downstream ColumnTransformer.
# experiments_dict['KDDCup09_appetency'] = {}
# experiments_dict['KDDCup09_appetency']['download_arff_url'] = 'https://www.openml.org/data/download/53994/KDDCup09_appetency.arff'
# experiments_dict['KDDCup09_appetency']['download_csv_url'] = 'https://www.openml.org/data/get_csv/53994/KDDCup09_appetency.arff'
# experiments_dict['KDDCup09_appetency']['task_type'] = 'classification'
# experiments_dict['KDDCup09_appetency']['target'] = 'appetency'

experiments_dict["Amazon_employee_access"] = {}
experiments_dict["Amazon_employee_access"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1681098/phpmPOD5A"
experiments_dict["Amazon_employee_access"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1681098/phpmPOD5A"
experiments_dict["Amazon_employee_access"]["task_type"] = "classification"
experiments_dict["Amazon_employee_access"]["target"] = "target"

experiments_dict["Fashion-MNIST"] = {}
experiments_dict["Fashion-MNIST"][
    "download_arff_url"
] = "https://www.openml.org/data/download/18238735/phpnBqZGZ"
experiments_dict["Fashion-MNIST"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/18238735/phpnBqZGZ"
experiments_dict["Fashion-MNIST"]["task_type"] = "classification"
experiments_dict["Fashion-MNIST"]["target"] = "class"

experiments_dict["dionis"] = {}
experiments_dict["dionis"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335690/file1c55272d7b5b.arff"
experiments_dict["dionis"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335690/file1c55272d7b5b.arff"
experiments_dict["dionis"]["task_type"] = "classification"
experiments_dict["dionis"]["target"] = "class"

experiments_dict["MiniBooNE"] = {}
experiments_dict["MiniBooNE"][
    "download_arff_url"
] = "https://www.openml.org/data/download/19335523/MiniBooNE.arff"
experiments_dict["MiniBooNE"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/19335523/MiniBooNE.arff"
experiments_dict["MiniBooNE"]["task_type"] = "classification"
experiments_dict["MiniBooNE"]["target"] = "signal"

experiments_dict["airlines"] = {}
experiments_dict["airlines"][
    "download_arff_url"
] = "https://www.openml.org/data/download/66526/phpvcoG8S"
experiments_dict["airlines"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/66526/phpvcoG8S"
experiments_dict["airlines"]["task_type"] = "stream classification"
experiments_dict["airlines"]["target"] = "class"

experiments_dict["diabetes"] = {}
experiments_dict["diabetes"]["dataset_url"] = "https://www.openml.org/d/37"
experiments_dict["diabetes"][
    "download_arff_url"
] = "https://www.openml.org/data/download/37/dataset_37_diabetes.arff"
experiments_dict["diabetes"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/37/dataset_37_diabetes.arff"
experiments_dict["diabetes"]["task_type"] = "classification"
experiments_dict["diabetes"]["target"] = "class"

experiments_dict["spectf"] = {}
experiments_dict["spectf"]["dataset_url"] = "https://www.openml.org/d/337"
experiments_dict["spectf"][
    "download_arff_url"
] = "https://www.openml.org/data/download/52240/phpDQbeeh"
experiments_dict["spectf"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/52240/phpDQbeeh"
experiments_dict["spectf"]["task_type"] = "classification"
experiments_dict["spectf"]["target"] = "overall_diagnosis"

experiments_dict["hill-valley"] = {}
experiments_dict["hill-valley"]["dataset_url"] = "https://www.openml.org/d/1479"
experiments_dict["hill-valley"][
    "download_arff_url"
] = "https://www.openml.org/data/download/1590101/php3isjYz"
experiments_dict["hill-valley"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/1590101/php3isjYz"
experiments_dict["hill-valley"]["task_type"] = "classification"
experiments_dict["hill-valley"]["target"] = "class"

experiments_dict["breast-cancer"] = {}
experiments_dict["breast-cancer"]["dataset_url"] = "https://www.openml.org/d/13"
experiments_dict["breast-cancer"][
    "download_arff_url"
] = "https://www.openml.org/data/download/13/dataset_13_breast-cancer.arff"
experiments_dict["breast-cancer"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/13/dataset_13_breast-cancer.arff"
experiments_dict["breast-cancer"]["task_type"] = "classification"
experiments_dict["breast-cancer"]["target"] = "class"

experiments_dict["compas"] = {}
experiments_dict["compas"][
    "download_arff_url"
] = "https://www.openml.org/data/download/21757035/compas.arff"
experiments_dict["compas"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/21757035/compas.arff"
experiments_dict["compas"]["task_type"] = "classification"
experiments_dict["compas"]["target"] = "two_year_recid"

experiments_dict["ricci"] = {}
experiments_dict["ricci"][
    "download_arff_url"
] = "https://www.openml.org/data/download/22044446/ricci_processed.arff"
experiments_dict["ricci"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/22044446/ricci_processed.arff"
experiments_dict["ricci"]["task_type"] = "classification"
experiments_dict["ricci"]["target"] = "promotion"

experiments_dict["SpeedDating"] = {}
experiments_dict["SpeedDating"]["dataset_url"] = "https://www.openml.org/d/40536"
experiments_dict["SpeedDating"][
    "download_arff_url"
] = "https://www.openml.org/data/download/13153954/speeddating.arff"
experiments_dict["SpeedDating"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/13153954/speeddating.arff"
experiments_dict["SpeedDating"]["task_type"] = "classification"
experiments_dict["SpeedDating"]["target"] = "match"

experiments_dict["nursery"] = {}
experiments_dict["nursery"]["dataset_url"] = "https://www.openml.org/d/26"
experiments_dict["nursery"][
    "download_arff_url"
] = "https://www.openml.org/data/download/26/dataset_26_nursery.arff"
experiments_dict["nursery"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/26/dataset_26_nursery.arff"
experiments_dict["nursery"]["task_type"] = "classification"
experiments_dict["nursery"]["target"] = "class"

experiments_dict["titanic"] = {}
experiments_dict["titanic"]["dataset_url"] = "https://www.openml.org/d/40945"
experiments_dict["titanic"][
    "download_arff_url"
] = "https://www.openml.org/data/download/16826755/phpMYEkMl"
experiments_dict["titanic"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/16826755/phpMYEkMl"
experiments_dict["titanic"]["task_type"] = "classification"
experiments_dict["titanic"]["target"] = "survived"

experiments_dict["tae"] = {}
experiments_dict["tae"]["dataset_url"] = "https://www.openml.org/d/48"
experiments_dict["tae"][
    "download_arff_url"
] = "https://www.openml.org/data/download/48/dataset_48_tae.arff"
experiments_dict["tae"][
    "download_csv_url"
] = "https://www.openml.org/data/get_csv/48/dataset_48_tae.arff"
experiments_dict["tae"]["task_type"] = "classification"
experiments_dict["tae"]["target"] = "class_attribute"


def add_schemas(schema_orig, target_col, train_X, test_X, train_y, test_y):
    from lale.datasets.data_schemas import add_schema

    elems_X = [
        item_schema
        for item_schema in schema_orig["items"]["items"]
        if item_schema["description"].lower() != target_col
    ]
    elem_y = [
        item_schema
        for item_schema in schema_orig["items"]["items"]
        if item_schema["description"].lower() == target_col
    ][0]
    if "enum" in elem_y:
        if isinstance(train_y, pd.Series):
            elem_y["enum"] = list(train_y.unique())
        else:
            elem_y["enum"] = [*range(len(elem_y["enum"]))]
    ncols_X = len(elems_X)
    rows_X = {
        **schema_orig["items"],
        "minItems": ncols_X,
        "maxItems": ncols_X,
        "items": elems_X,
    }
    if "json_schema" not in pd.DataFrame._internal_names:
        pd.DataFrame._internal_names.append("json_schema")
    nrows_train, nrows_test = len(train_y), len(test_y)
    train_X = add_schema(
        train_X,
        {
            **schema_orig,
            "minItems": nrows_train,
            "maxItems": nrows_train,
            "items": rows_X,
        },
    )
    test_X = add_schema(
        test_X,
        {
            **schema_orig,
            "minItems": nrows_test,
            "maxItems": nrows_test,
            "items": rows_X,
        },
    )
    train_y = add_schema(
        train_y,
        {
            **schema_orig,
            "minItems": nrows_train,
            "maxItems": nrows_train,
            "items": elem_y,
        },
    )
    test_y = add_schema(
        test_y,
        {
            **schema_orig,
            "minItems": nrows_test,
            "maxItems": nrows_test,
            "items": elem_y,
        },
    )
    return train_X, test_X, train_y, test_y


numeric_data_types_list = ["numeric", "integer", "real"]


def fetch(
    dataset_name, task_type, verbose=False, preprocess=True, test_size=0.33, astype=None
):
    if verbose:
        print("Loading dataset:", dataset_name)
    # Check that the dataset name exists in experiments_dict
    try:
        if experiments_dict[dataset_name]["task_type"] != task_type.lower():
            raise ValueError(
                "The task type {} does not match with the given datasets task type {}".format(
                    task_type, experiments_dict[dataset_name]["task_type"]
                )
            )

    except KeyError:
        raise KeyError(
            "Dataset name {} not found in the supported datasets".format(dataset_name)
        )
    data_file_name = os.path.join(download_data_dir, dataset_name + ".arff")
    if verbose:
        print(f"data file name: {data_file_name}")
    if not os.path.exists(data_file_name):
        # TODO: Download the data
        if not os.path.exists(download_data_dir):
            os.makedirs(download_data_dir)
            if verbose:
                print("created directory {}".format(download_data_dir))
        urllib.request.urlretrieve(
            experiments_dict[dataset_name]["download_arff_url"], data_file_name
        )

    assert os.path.exists(data_file_name)
    with open(data_file_name) as f:
        dataDictionary = arff.load(f)
        f.close()

    from lale.datasets.data_schemas import liac_arff_to_schema

    schema_orig = liac_arff_to_schema(dataDictionary)
    target_col = experiments_dict[dataset_name]["target"]
    y: Optional[Any] = None
    if preprocess:
        arffData = pd.DataFrame(dataDictionary["data"])
        # arffData = arffData.fillna(0)
        attributes = dataDictionary["attributes"]

        if verbose:
            print(f"attributes: {attributes}")
        categorical_cols = []
        numeric_cols = []
        X_columns = []
        for i, item in enumerate(attributes):
            if item[0].lower() == target_col:
                target_indx = i
                # remove it from attributes so that the next loop indices are adjusted accordingly.
                del attributes[i]
                # the type stubs for pandas are not currently complete enough to type this correctly
                y = arffData.iloc[:, target_indx]  # type: ignore
                arffData = arffData.drop(i, axis=1)

        for i, item in enumerate(attributes):
            X_columns.append(i)
            if (
                (
                    isinstance(item[1], str)
                    and item[1].lower() not in numeric_data_types_list
                )
                or isinstance(item[1], list)
            ) and (item[0].lower() != "class"):
                categorical_cols.append(i)
            elif (
                isinstance(item[1], str) and item[1].lower() in numeric_data_types_list
            ) and (item[0].lower() != "class"):
                numeric_cols.append(i)
        if verbose:
            print(f"categorical columns: {categorical_cols}")
            print(f"numeric columns:     {numeric_cols}")
        X = arffData.iloc[:, X_columns]

        # Check whether there is any error
        num_classes_from_last_row = len(list(set(y))) if y is not None else 0

        if verbose:
            print("num_classes_from_last_row", num_classes_from_last_row)

        transformers1 = [
            (
                "imputer_str",
                SimpleImputer(missing_values=None, strategy="most_frequent"),
                categorical_cols,
            ),
            ("imputer_num", SimpleImputer(strategy="mean"), numeric_cols),
        ]
        txm1 = ColumnTransformer(transformers1, sparse_threshold=0.0)

        transformers2 = [
            ("ohe", OneHotEncoder(sparse=False), list(range(len(categorical_cols)))),
            (
                "no_op",
                "passthrough",
                list(
                    range(
                        len(categorical_cols), len(categorical_cols) + len(numeric_cols)
                    )
                ),
            ),
        ]
        txm2 = ColumnTransformer(transformers2, sparse_threshold=0.0)
        if verbose:
            print("Shape of X before preprocessing", X.shape)
        from sklearn.pipeline import make_pipeline

        preprocessing = make_pipeline(txm1, txm2)

        X = preprocessing.fit(X).transform(X)

        if verbose:
            print(f"shape of X after preprocessing: {X.shape}")

        if astype in ["pandas", "spark"]:
            cat_col_names = [attributes[i][0].lower() for i in categorical_cols]
            one_hot_encoder = preprocessing.steps[1][1].named_transformers_["ohe"]
            encoded_names = one_hot_encoder.get_feature_names(cat_col_names)
            num_col_names = [attributes[i][0].lower() for i in numeric_cols]
            col_names = list(encoded_names) + list(num_col_names)
            if verbose:
                print(f"column names after preprocessing: {col_names}")
            X = pd.DataFrame(X, columns=col_names)

    else:
        col_names = [attr[0].lower() for attr in dataDictionary["attributes"]]
        df_all = pd.DataFrame(dataDictionary["data"], columns=col_names)
        y = df_all[target_col]
        # the type stubs for pandas are not currently complete enough to type this correctly
        y = y.squeeze()  # type: ignore
        cols_X = [col for col in col_names if col != target_col]
        X = df_all[cols_X]

    if preprocess:
        labelencoder = LabelEncoder()
        y = labelencoder.fit_transform(y)
    if astype in ["pandas", "spark"] and not isinstance(y, pd.Series):
        y = pd.Series(y, name=target_col)

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=0
    )
    if verbose:
        print(f"training set shapes: X {X_train.shape}, y {y_train.shape}")
        print(f"test set shapes:     X {X_test.shape}, y {y_test.shape}")
    if preprocess:
        from lale.datasets.data_schemas import add_schema

        X_train = add_schema(X_train.astype(np.number), recalc=True)
        y_train = add_schema(y_train.astype(int), recalc=True)
        X_test = add_schema(X_test.astype(np.number), recalc=True)
        y_test = add_schema(y_test.astype(int), recalc=True)
    else:
        X_train, X_test, y_train, y_test = add_schemas(
            schema_orig, target_col, X_train, X_test, y_train, y_test
        )
    if astype == "spark":
        from lale.datasets import pandas2spark

        X_train = pandas2spark(X_train)
        X_test = pandas2spark(X_test)

    return (X_train, y_train), (X_test, y_test)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Schema-enhanced versions of some of the operators from `scikit-learn`_ to enable hyperparameter tuning.

.. _`scikit-learn`: https://scikit-learn.org

Operators
=========

Classifiers:

* lale.lib.sklearn. `AdaBoostClassifier`_
* lale.lib.sklearn. `BaggingClassifier`_
* lale.lib.sklearn. `DecisionTreeClassifier`_
* lale.lib.sklearn. `DummyClassifier`_
* lale.lib.sklearn. `ExtraTreesClassifier`_
* lale.lib.sklearn. `GaussianNB`_
* lale.lib.sklearn. `GradientBoostingClassifier`_
* lale.lib.sklearn. `KNeighborsClassifier`_
* lale.lib.sklearn. `LinearSVC`_
* lale.lib.sklearn. `LogisticRegression`_
* lale.lib.sklearn. `MLPClassifier`_
* lale.lib.sklearn. `MultinomialNB`_
* lale.lib.sklearn. `PassiveAggressiveClassifier`_
* lale.lib.sklearn. `RandomForestClassifier`_
* lale.lib.sklearn. `RidgeClassifier`_
* lale.lib.sklearn. `SGDClassifier`_
* lale.lib.sklearn. `StackingClassifier`_
* lale.lib.sklearn. `SVC`_
* lale.lib.sklearn. `VotingClassifier`_

Regressors:

* lale.lib.sklearn. `AdaBoostRegressor`_
* lale.lib.sklearn. `BaggingRegressor`_
* lale.lib.sklearn. `DecisionTreeRegressor`_
* lale.lib.sklearn. `DummyRegressor`_
* lale.lib.sklearn. `ExtraTreesRegressor`_
* lale.lib.sklearn. `GradientBoostingRegressor`_
* lale.lib.sklearn. `KNeighborsRegressor`_
* lale.lib.sklearn. `LinearRegression`_
* lale.lib.sklearn. `LinearSVR`_
* lale.lib.sklearn. `MultiOutputRegressor`_
* lale.lib.sklearn. `RandomForestRegressor`_
* lale.lib.sklearn. `Ridge`_
* lale.lib.sklearn. `SGDRegressor`_
* lale.lib.sklearn. `StackingRegressor`_
* lale.lib.sklearn. `SVR`_
* lale.lib.sklearn. `VotingRegressor`_

Transformers:

* lale.lib.sklearn. `ColumnTransformer`_
* lale.lib.sklearn. `FeatureAgglomeration`_
* lale.lib.sklearn. `FunctionTransformer`_
* lale.lib.sklearn. `IsolationForest`_
* lale.lib.sklearn. `Isomap`_
* lale.lib.sklearn. `MinMaxScaler`_
* lale.lib.sklearn. `MissingIndicator`_
* lale.lib.sklearn. `NMF`_
* lale.lib.sklearn. `Normalizer`_
* lale.lib.sklearn. `Nystroem`_
* lale.lib.sklearn. `OneHotEncoder`_
* lale.lib.sklearn. `OrdinalEncoder`_
* lale.lib.sklearn. `PCA`_
* lale.lib.sklearn. `PolynomialFeatures`_
* lale.lib.sklearn. `QuadraticDiscriminantAnalysis`_
* lale.lib.sklearn. `QuantileTransformer`_
* lale.lib.sklearn. `RFE`_
* lale.lib.sklearn. `RobustScaler`_
* lale.lib.sklearn. `SelectKBest`_
* lale.lib.sklearn. `SimpleImputer`_
* lale.lib.sklearn. `StandardScaler`_
* lale.lib.sklearn. `TfidfVectorizer`_
* lale.lib.sklearn. `VarianceThreshold`_

Estimators and transformers:

* lale.lib.sklearn. `Pipeline`_

Clustering:

* lale.lib.sklearn. `KMeans`_

.. _`AdaBoostClassifier`: lale.lib.sklearn.ada_boost_classifier.html
.. _`AdaBoostRegressor`: lale.lib.sklearn.ada_boost_regressor.html
.. _`BaggingClassifier`: lale.lib.sklearn.bagging_classifier.html
.. _`BaggingRegressor`: lale.lib.sklearn.bagging_regressor.html
.. _`ColumnTransformer`: lale.lib.sklearn.column_transformer.html
.. _`DecisionTreeClassifier`: lale.lib.sklearn.decision_tree_classifier.html
.. _`DecisionTreeRegressor`: lale.lib.sklearn.decision_tree_regressor.html
.. _`DummyClassifier`: lale.lib.sklearn.dummy_classifier.html
.. _`DummyRegressor`: lale.lib.sklearn.dummy_regressor.html
.. _`ExtraTreesClassifier`: lale.lib.sklearn.extra_trees_classifier.html
.. _`ExtraTreesRegressor`: lale.lib.sklearn.extra_trees_regressor.html
.. _`FeatureAgglomeration`: lale.lib.sklearn.feature_agglomeration.html
.. _`FunctionTransformer`: lale.lib.sklearn.function_transformer.html
.. _`GaussianNB`: lale.lib.sklearn.gaussian_nb.html
.. _`GradientBoostingClassifier`: lale.lib.sklearn.gradient_boosting_classifier.html
.. _`GradientBoostingRegressor`: lale.lib.sklearn.gradient_boosting_regressor.html
.. _`IsolationForest`: lale.lib.sklearn.isolation_forest.html
.. _`Isomap`: lale.lib.sklearn.isomap.html
.. _`KMeans`: lale.lib.sklearn.k_means.html
.. _`KNeighborsClassifier`: lale.lib.sklearn.k_neighbors_classifier.html
.. _`KNeighborsRegressor`: lale.lib.sklearn.k_neighbors_regressor.html
.. _`LinearRegression`: lale.lib.sklearn.linear_regression.html
.. _`LinearSVC`: lale.lib.sklearn.linear_svc.html
.. _`LogisticRegression`: lale.lib.sklearn.logistic_regression.html
.. _`MinMaxScaler`: lale.lib.sklearn.min_max_scaler.html
.. _`MissingIndicator`: lale.lib.sklearn.missing_indicator.html
.. _`MLPClassifier`: lale.lib.sklearn.mlp_classifier.html
.. _`MultinomialNB`: lale.lib.sklearn.multinomial_nb.html
.. _`MultiOutputRegressor`: lale.lib.sklearn.multi_output_regressor.html
.. _`NMF`: lale.lib.sklearn.nmf.html
.. _`Normalizer`: lale.lib.sklearn.normalizer.html
.. _`Nystroem`: lale.lib.sklearn.nystroem.html
.. _`OneHotEncoder`: lale.lib.sklearn.one_hot_encoder.html
.. _`OrdinalEncoder`: lale.lib.sklearn.ordinal_encoder.html
.. _`PassiveAggressiveClassifier`: lale.lib.sklearn.passive_aggressive_classifier.html
.. _`PCA`: lale.lib.sklearn.pca.html
.. _`Pipeline`: lale.lib.sklearn.pipeline.html
.. _`PolynomialFeatures`: lale.lib.sklearn.polynomial_features.html
.. _`QuadraticDiscriminantAnalysis`: lale.lib.sklearn.quadratic_discriminant_analysis.html
.. _`QuantileTransformer`: lale.lib.sklearn.quantile_transformer.html
.. _`RandomForestClassifier`: lale.lib.sklearn.random_forest_classifier.html
.. _`RandomForestRegressor`: lale.lib.sklearn.random_forest_regressor.html
.. _`RFE`: lale.lib.sklearn.rfe.html
.. _`Ridge`: lale.lib.sklearn.ridge.html
.. _`RidgeClassifier`: lale.lib.sklearn.ridge_classifier.html
.. _`RobustScaler`: lale.lib.sklearn.robust_scaler.html
.. _`SelectKBest`: lale.lib.sklearn.select_k_best.html
.. _`SGDClassifier`: lale.lib.sklearn.sgd_classifier.html
.. _`SGDRegressor`: lale.lib.sklearn.sgd_regressor.html
.. _`SimpleImputer`: lale.lib.sklearn.simple_imputer.html
.. _`StandardScaler`: lale.lib.sklearn.standard_scaler.html
.. _`SVC`: lale.lib.sklearn.svc.html
.. _`SVR`: lale.lib.sklearn.svr.html
.. _`LinearSVR`: lale.lib.sklearn.linear_svr.html
.. _`TfidfVectorizer`: lale.lib.sklearn.tfidf_vectorizer.html
.. _`VarianceThreshold`: lale.lib.sklearn.variance_threshold.html
.. _`VotingClassifier`: lale.lib.sklearn.voting_classifier.html
.. _`VotingRegressor`: lale.lib.sklearn.voting_regressor.html
.. _`StackingClassifier`: lale.lib.sklearn.stacking_classifier.html
.. _`StackingRegressor`: lale.lib.sklearn.stacking_regressor.html
"""

from sklearn import __version__ as sklearn_version

from .ada_boost_classifier import AdaBoostClassifier
from .ada_boost_regressor import AdaBoostRegressor
from .bagging_classifier import BaggingClassifier
from .bagging_regressor import BaggingRegressor
from .column_transformer import ColumnTransformer
from .decision_tree_classifier import DecisionTreeClassifier
from .decision_tree_regressor import DecisionTreeRegressor
from .dummy_classifier import DummyClassifier
from .dummy_regressor import DummyRegressor
from .extra_trees_classifier import ExtraTreesClassifier
from .extra_trees_regressor import ExtraTreesRegressor
from .feature_agglomeration import FeatureAgglomeration
from .function_transformer import FunctionTransformer
from .gaussian_nb import GaussianNB
from .gradient_boosting_classifier import GradientBoostingClassifier
from .gradient_boosting_regressor import GradientBoostingRegressor
from .isolation_forest import IsolationForest
from .isomap import Isomap
from .k_means import KMeans
from .k_neighbors_classifier import KNeighborsClassifier
from .k_neighbors_regressor import KNeighborsRegressor
from .linear_regression import LinearRegression
from .linear_svc import LinearSVC
from .linear_svr import LinearSVR
from .logistic_regression import LogisticRegression
from .min_max_scaler import MinMaxScaler
from .missing_indicator import MissingIndicator
from .mlp_classifier import MLPClassifier
from .multi_output_regressor import MultiOutputRegressor
from .multinomial_nb import MultinomialNB
from .nmf import NMF
from .normalizer import Normalizer
from .nystroem import Nystroem
from .one_hot_encoder import OneHotEncoder
from .ordinal_encoder import OrdinalEncoder
from .passive_aggressive_classifier import PassiveAggressiveClassifier
from .pca import PCA
from .pipeline import Pipeline
from .polynomial_features import PolynomialFeatures
from .quadratic_discriminant_analysis import QuadraticDiscriminantAnalysis
from .quantile_transformer import QuantileTransformer
from .random_forest_classifier import RandomForestClassifier
from .random_forest_regressor import RandomForestRegressor
from .rfe import RFE
from .ridge import Ridge
from .ridge_classifier import RidgeClassifier
from .robust_scaler import RobustScaler
from .select_k_best import SelectKBest
from .sgd_classifier import SGDClassifier
from .sgd_regressor import SGDRegressor
from .simple_imputer import SimpleImputer

if sklearn_version >= "0.21":
    from .stacking_classifier import StackingClassifier
    from .stacking_regressor import StackingRegressor
    from .voting_regressor import VotingRegressor

from .standard_scaler import StandardScaler
from .svc import SVC
from .svr import SVR
from .tfidf_vectorizer import TfidfVectorizer
from .variance_threshold import VarianceThreshold
from .voting_classifier import VotingClassifier



import sklearn
from numpy import inf, nan
from sklearn.calibration import CalibratedClassifierCV as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _CalibratedClassifierCVImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for CalibratedClassifierCV    Probability calibration with isotonic regression or sigmoid.",
    "allOf": [
        {
            "type": "object",
            "required": ["base_estimator", "method", "cv"],
            "relevantToOptimizer": ["method", "cv"],
            "additionalProperties": False,
            "properties": {
                "base_estimator": {
                    "XXX TODO XXX": "instance BaseEstimator",
                    "description": "The classifier whose output decision function needs to be calibrated to offer more accurate predict_proba outputs",
                    "enum": [None],
                    "default": None,
                },
                "method": {
                    "XXX TODO XXX": "'sigmoid' or 'isotonic'",
                    "description": "The method to use for calibration",
                    "enum": ["sigmoid", "isotonic"],
                    "default": "sigmoid",
                },
                "cv": {
                    "description": """Cross-validation as integer or as object that has a split function.
                        The fit method performs cross validation on the input dataset for per
                        trial, and uses the mean cross validation performance for optimization.
                        This behavior is also impacted by handle_cv_failure flag.
                        If integer: number of folds in sklearn.model_selection.StratifiedKFold.
                        If object with split function: generator yielding (train, test) splits
                        as arrays of indices. Can use any of the iterators from
                        https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators.""",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "default": 5,
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 4,
                            "distribution": "uniform",
                        },
                        {"laleType": "Any", "forOptimizer": False},
                        {"enum": [None]},
                        {"enum": ["prefit"]},
                    ],
                    "default": None,
                },
            },
        }
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the calibrated model",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data.",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values.",
        },
        "sample_weight": {
            "anyOf": [{"type": "array", "items": {"type": "number"}}, {"enum": [None]}],
            "description": "Sample weights",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict the target of new samples. Can be different from the",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "The samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "The predicted class.",
    "type": "array",
    "items": {"type": "number"},
}
_input_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Posterior probabilities of classification",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "The samples.",
        }
    },
}
_output_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "The predicted probas.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.calibration.CalibratedClassifierCV#sklearn-calibration-calibratedclassifiercv",
    "import_from": "sklearn.calibration",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}
CalibratedClassifierCV = make_operator(_CalibratedClassifierCVImpl, _combined_schemas)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.calibration.CalibratedClassifierCV#sklearn-calibration-calibratedclassifiercv
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.calibration.CalibratedClassifierCV#sklearn-calibration-calibratedclassifiercv
    CalibratedClassifierCV = CalibratedClassifierCV.customize_schema(
        n_jobs={
            "description": "Number of jobs to run in parallel.",
            "anyOf": [
                {
                    "description": "1 unless in joblib.parallel_backend context.",
                    "enum": [None],
                },
                {"description": "Use all processors.", "enum": [-1]},
                {
                    "description": "Number of jobs to run in parallel.",
                    "type": "integer",
                    "minimum": 1,
                },
            ],
            "default": None,
        },
        set_as_available=True,
    )
    CalibratedClassifierCV = CalibratedClassifierCV.customize_schema(
        ensemble={
            "type": "boolean",
            "default": True,
            "description": "Determines how the calibrator is fitted when cv is not 'prefit'. Ignored if cv='prefit",
        },
        set_as_available=True,
    )


set_docstrings(CalibratedClassifierCV)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch.utils.data import Dataset


class BatchDataDict(Dataset):
    """Pytorch Dataset subclass that takes a dictionary of format {'<batch_idx>': <batch_data>}."""

    def __init__(self, X, y=None):
        """X is the dictionary dataset and y is ignored.

        Parameters
        ----------
        X : dict
         Dictionary of format {'<batch_idx>': <batch_data>}
        y : None
            Ignored.
        """
        self.data_dict = X

    def __len__(self):
        return len(self.data_dict)

    def __getitem__(self, idx):
        # This returns the batch at idx instead of a single element.
        return self.data_dict[idx]



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Any

import numpy as np

try:
    import pyspark.sql.functions

    spark_installed = True
except ImportError:
    spark_installed = False


from lale.helpers import _is_pandas_df, _is_spark_df


class categorical:
    """Creates a callable for projecting categorical columns with sklearn's ColumnTransformer or Lale's Project operator.

    Parameters
    ----------
    max_values : int

        Maximum number of unique values in a column for it to be considered categorical.

    Returns
    -------
    callable
        Function that, given a dataset X, returns a list of columns,
        containing either string column names or integer column indices."""

    def __init__(self, max_values: int = 5):
        self._max_values = max_values

    def __repr__(self):
        return f"lale.lib.lale.categorical(max_values={self._max_values})"

    def __call__(self, X):
        def is_categorical(column_values):
            unique_values = set()
            for val in column_values:
                if val not in unique_values:
                    unique_values.add(val)
                    if len(unique_values) > self._max_values:
                        return False
            return True

        if _is_pandas_df(X):
            result = [c for c in X.columns if is_categorical(X[c])]
        elif isinstance(X, np.ndarray):
            result = [c for c in range(X.shape[1]) if is_categorical(X[:, c])]
        else:
            raise TypeError(f"unexpected type {type(X)}")
        return result


class date_time:
    """Creates a callable for projecting date/time columns with sklearn's ColumnTransformer or Lale's Project operator.

    Parameters
    ----------
    fmt : str

        Format string for `strptime()`, see https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior

    Returns
    -------
    callable
        Function that, given a dataset X, returns a list of columns,
        containing either string column names or integer column indices."""

    def __init__(self, fmt):
        self._fmt = fmt

    def __repr__(self):
        return f"lale.lib.lale.date_time(fmt={self._fmt})"

    def __call__(self, X):
        def is_date_time(column_values):
            try:
                for val in column_values:
                    if isinstance(val, str):
                        datetime.datetime.strptime(val, self._fmt)
                    else:
                        return False
            except ValueError:
                return False
            return True

        if _is_pandas_df(X):
            result = [c for c in X.columns if is_date_time(X[c])]
        elif isinstance(X, np.ndarray):
            result = [c for c in range(X.shape[1]) if is_date_time(X[:, c])]
        else:
            raise TypeError(f"unexpected type {type(X)}")
        return result


# functions for filter
def filter_isnan(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].isnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(pyspark.sql.functions.isnan(df[column_name]))
    else:
        raise ValueError(
            "the filter isnan supports only Pandas dataframes or spark dataframes."
        )


def filter_isnotnan(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].notnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(~pyspark.sql.functions.isnan(df[column_name]))
    else:
        raise ValueError(
            "the filter isnotnan supports only Pandas dataframes or spark dataframes."
        )


def filter_isnull(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].isnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(pyspark.sql.functions.isnull(df[column_name]))
    else:
        raise ValueError(
            "the filter isnan supports only Pandas dataframes or spark dataframes."
        )


def filter_isnotnull(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].notnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(~pyspark.sql.functions.isnull(df[column_name]))
    else:
        raise ValueError(
            "the filter isnotnan supports only Pandas dataframes or spark dataframes."
        )



from numpy import inf, nan
from sklearn.linear_model import TheilSenRegressor as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _TheilSenRegressorImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for TheilSenRegressor    Theil-Sen Estimator: robust multivariate regression model.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "fit_intercept",
                "copy_X",
                "max_subpopulation",
                "n_subsamples",
                "max_iter",
                "tol",
                "random_state",
                "n_jobs",
                "verbose",
            ],
            "relevantToOptimizer": [
                "fit_intercept",
                "copy_X",
                "max_subpopulation",
                "max_iter",
                "tol",
            ],
            "additionalProperties": False,
            "properties": {
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to calculate the intercept for this model",
                },
                "copy_X": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, X will be copied; else, it may be overwritten.",
                },
                "max_subpopulation": {
                    "type": "integer",
                    "minimumForOptimizer": 10000,
                    "maximumForOptimizer": 10001,
                    "distribution": "uniform",
                    "default": 10000,
                    "description": "Instead of computing with a set of cardinality 'n choose k', where n is the number of samples and k is the number of subsamples (at least number of features), consider only a stochastic subpopulation of a given maximal size if 'n choose k' is larger than max_subpopulation",
                },
                "n_subsamples": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": None,
                    "description": "Number of samples to calculate the parameters",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 300,
                    "description": "Maximum number of iterations for the calculation of spatial median.",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.001,
                    "description": "Tolerance when calculating spatial median.",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "A random number generator instance to define the state of the random permutations generator",
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": 1,
                    "description": "Number of CPUs to use during the cross validation",
                },
                "verbose": {
                    "type": "boolean",
                    "default": False,
                    "description": "Verbose mode when fitting the model.",
                },
            },
        },
        {
            "XXX TODO XXX": "Parameter: max_subpopulation > only a stochastic subpopulation of a given maximal size if 'n choose k' is larger than max_subpopulation"
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit linear model.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the linear model",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Returns predicted values.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.TheilSenRegressor#sklearn-linear_model-theilsenregressor",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
TheilSenRegressor = make_operator(_TheilSenRegressorImpl, _combined_schemas)

set_docstrings(TheilSenRegressor)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
from sklearn.feature_selection import RFE as SKLModel

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Feature ranking with recursive feature elimination.",
    "allOf": [
        {
            "type": "object",
            "required": ["estimator", "n_features_to_select", "step", "verbose"],
            "relevantToOptimizer": [],
            "additionalProperties": False,
            "properties": {
                "estimator": {
                    "description": "A supervised learning estimator with a fit method that provides information about feature importance either through a `coef_` attribute or through a `feature_importances_` attribute.",
                    "laleType": "operator",
                },
                "n_features_to_select": {
                    "description": "The number of features to select. If None, half of the features are selected.",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "laleMaximum": "X/items/maxItems",  # number of columns
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "step": {
                    "description": "If greater than or equal to 1, then step corresponds to the (integer) number of features to remove at each iteration. If within (0.0, 1.0), then step corresponds to the percentage (rounded down) of features to remove at each iteration.",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "forOptimizer": False,
                        },
                        {
                            "type": "number",
                            "minimum": 0,
                            "exclusiveMinimum": True,
                            "maximum": 1,
                            "exclusiveMaximum": True,
                            "maximumForOptimizer": 0.5,
                        },
                    ],
                    "default": 1,
                },
                "verbose": {
                    "anyOf": [{"type": "boolean"}, {"type": "integer"}],
                    "default": 0,
                    "description": "Controls verbosity of output.",
                },
            },
        }
    ],
}

_input_fit_schema = {
    "description": "Fit the model to data matrix X and target(s) y.",
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        "y": {
            "description": "Target class labels; the array is over samples.",
            "type": "array",
            "items": {"type": "number"},
        },
    },
}

_input_transform_schema = {
    "description": "Reduce X to the selected features.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The input samples.",
        }
    },
}

_output_transform_schema = {
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Recursive feature elimination`_ transformer from scikit-learn.

.. _`Recursive feature elimination`: https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFE.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.rfe.html",
    "import_from": "sklearn.feature_selection",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

RFE: lale.operators.PlannedIndividualOp
RFE = lale.operators.make_operator(SKLModel, _combined_schemas)


if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.feature_selection.RFE.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.feature_selection.RFE.html
    RFE = RFE.customize_schema(
        n_features_to_select={
            "anyOf": [
                {
                    "description": "Half of the features are selected.",
                    "enum": [None],
                },
                {
                    "description": "Absolute number of features to select.",
                    "type": "integer",
                    "minimum": 1,
                    "laleMaximum": "X/items/maxItems",  # number of columns
                    "forOptimizer": False,
                },
                {
                    "description": "Fraction of features to select",
                    "type": "number",
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "maximum": 1.0,
                    "exclusiveMaximum": True,
                },
            ],
            "default": None,
        },
        importance_getter={
            "anyOf": [
                {
                    "description": "Use the feature importance either through a `coef_` or `feature_importances_` attributes of estimator.",
                    "enum": ["auto"],
                },
                {
                    "description": "Attribute name/path for extracting feature importance (implemented with attrgetter).",
                    "type": "string",
                },
                {
                    "description": "The callable is passed with the fitted estimator and it should return importance for each feature.",
                    "laleType": "callable",
                },
            ],
            "default": "auto",
        },
        set_as_available=True,
    )


lale.docstrings.set_docstrings(RFE)



from numpy import inf, nan
from sklearn.manifold import Isomap as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _IsomapImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for Isomap    Isomap Embedding",
    "allOf": [
        {
            "type": "object",
            "required": [
                "n_neighbors",
                "n_components",
                "eigen_solver",
                "tol",
                "max_iter",
                "path_method",
                "neighbors_algorithm",
                "n_jobs",
            ],
            "relevantToOptimizer": [
                "n_neighbors",
                "n_components",
                "eigen_solver",
                "tol",
                "path_method",
                "neighbors_algorithm",
            ],
            "additionalProperties": False,
            "properties": {
                "n_neighbors": {
                    "type": "integer",
                    "minimumForOptimizer": 5,
                    "maximumForOptimizer": 20,
                    "distribution": "uniform",
                    "default": 5,
                    "description": "number of neighbors to consider for each point.",
                },
                "n_components": {
                    "type": "integer",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 256,
                    "distribution": "uniform",
                    "default": 2,
                    "description": "number of coordinates for the manifold",
                },
                "eigen_solver": {
                    "enum": ["auto", "arpack", "dense"],
                    "default": "auto",
                    "description": "'auto' : Attempt to choose the most efficient solver for the given problem",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 0,
                    "maximumForOptimizer": 1,
                    "distribution": "uniform",
                    "default": 0,
                    "description": "Convergence tolerance passed to arpack or lobpcg",
                },
                "max_iter": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": None,
                    "description": "Maximum number of iterations for the arpack solver",
                },
                "path_method": {
                    "enum": ["auto", "FW", "D"],
                    "default": "auto",
                    "description": "Method to use in finding shortest path",
                },
                "neighbors_algorithm": {
                    "enum": ["auto", "brute", "kd_tree", "ball_tree"],
                    "default": "auto",
                    "description": "Algorithm to use for nearest neighbors search, passed to neighbors.NearestNeighbors instance.",
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": 1,
                    "description": "The number of parallel jobs to run",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Compute the embedding vectors for data X",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
            "XXX TODO XXX": "{array-like, sparse matrix, BallTree, KDTree, NearestNeighbors}",
            "description": "Sample data, shape = (n_samples, n_features), in the form of a numpy array, precomputed tree, or NearestNeighbors object.",
        },
        "y": {},
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Transform X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}}
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Transform X.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.manifold.Isomap#sklearn-manifold-isomap",
    "import_from": "sklearn.manifold",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}
Isomap = make_operator(_IsomapImpl, _combined_schemas)

set_docstrings(Isomap)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.preprocessing

import lale.docstrings
import lale.operators

_input_schema_fit = {
    "description": "Input data schema for training.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        "y": {},
    },
}

_input_transform_schema = {
    "description": "Input data schema for predictions.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        }
    },
}

_output_transform_schema = {
    "description": "Output data schema for transformed data.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "additionalProperties": False,
            "required": ["feature_range", "copy"],
            "relevantToOptimizer": [],  # ['feature_range'],
            "properties": {
                "feature_range": {
                    "description": "Desired range of transformed data.",
                    "type": "array",
                    "laleType": "tuple",
                    "minItems": 2,
                    "maxItems": 2,
                    "items": [
                        {
                            "type": "number",
                            "minimumForOptimizer": -1,
                            "maximumForOptimizer": 0,
                        },
                        {
                            "type": "number",
                            "minimumForOptimizer": 0.001,
                            "maximumForOptimizer": 1,
                        },
                    ],
                    "default": (0, 1),
                },
                "copy": {
                    "description": "Set to False to perform inplace row normalization and avoid "
                    "a copy (if the input is already a numpy array).",
                    "type": "boolean",
                    "default": True,
                },
            },
        },
        {
            "description": "MinMaxScaler does not support sparse input. Consider using MaxAbsScaler instead.",
            "type": "object",
            "laleNot": "X/isSparse",
        },
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Min-max scaler`_ transformer from scikit-learn.

.. _`Min-max scaler`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.min_max_scaler.html",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {
        "pre": ["~categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_schema_fit,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

MinMaxScaler: lale.operators.PlannedIndividualOp
MinMaxScaler = lale.operators.make_operator(
    sklearn.preprocessing.MinMaxScaler, _combined_schemas
)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.22/modules/generated/sklearn.preprocessing.MinMaxScaler.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.preprocessing.MinMaxScaler.html
    MinMaxScaler = MinMaxScaler.customize_schema(
        clip={
            "type": "boolean",
            "description": "Set to True to clip transformed values of held-out data to provided feature range.",
            "default": False,
        },
        set_as_available=True,
    )


lale.docstrings.set_docstrings(MinMaxScaler)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
from sklearn.cluster import KMeans as SKLModel

from lale.docstrings import set_docstrings
from lale.operators import make_operator

_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """The k-means problem is solved using either Lloyd's or Elkan's algorithm.
The average complexity is given by O(k n T), where n is the number of
samples and T is the number of iteration.
The worst case complexity is given by O(n^(k+2/p)) with
n = n_samples, p = n_features. (D. Arthur and S. Vassilvitskii,
'How slow is the k-means method?' SoCG2006)
In practice, the k-means algorithm is very fast (one of the fastest
clustering algorithms available), but it falls in local minima. That's why
it can be useful to restart it several times.
If the algorithm stops before fully converging (because of ``tol`` or
``max_iter``), ``labels_`` and ``cluster_centers_`` will not be consistent,
i.e. the ``cluster_centers_`` will not be the means of the points in each
cluster. Also, the estimator will reassign ``labels_`` after the last
iteration to make ``labels_`` consistent with ``predict`` on the training
set.""",
    "allOf": [
        {
            "type": "object",
            "required": [
                "n_clusters",
                "init",
                "n_init",
                "max_iter",
                "tol",
                "precompute_distances",
                "verbose",
                "random_state",
                "copy_x",
                "n_jobs",
                "algorithm",
            ],
            "relevantToOptimizer": [
                "n_clusters",
                "init",
                "n_init",
                "max_iter",
                "tol",
                "precompute_distances",
                "copy_x",
                "algorithm",
            ],
            "additionalProperties": False,
            "properties": {
                "n_clusters": {
                    "type": "integer",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 8,
                    "distribution": "uniform",
                    "default": 8,
                    "description": "The number of clusters to form as well as the number of centroids to generate.",
                },
                "init": {
                    "anyOf": [
                        {"enum": ["k-means++", "random"]},
                        {"laleType": "callable", "forOptimizer": False},
                        {
                            "type": "array",
                            "items": {"type": "array", "items": {"type": "number"}},
                            "forOptimizer": False,
                        },
                    ],
                    "default": "k-means++",
                    "description": """Method for initialization, defaults to `k-means++`.
`k-means++` : selects initial cluster centers for k-mean clustering in a smart way to speed up convergence.
See section Notes in k_init for more details.
`random`: choose n_clusters observations (rows) at random from data for the initial centroids.
If an array is passed, it should be of shape (n_clusters, n_features) and gives the initial centers.
If a callable is passed, it should take arguments X, n_clusters and a random state and return an initialization.""",
                },
                "n_init": {
                    "type": "integer",
                    "minimumForOptimizer": 3,
                    "maximumForOptimizer": 10,
                    "distribution": "uniform",
                    "default": 10,
                    "description": """Number of time the k-means algorithm will be run with different centroid seeds.
The final results will be the best output of n_init consecutive runs in terms of inertia.""",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 300,
                    "description": "Maximum number of iterations of the k-means algorithm for a single run.",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "Relative tolerance with regards to Frobenius norm of the difference in the cluster centers of two consecutive iterations to declare convergence.",
                },
                "precompute_distances": {
                    "enum": ["auto", True, False],
                    "default": "auto",
                    "description": "Precompute distances (faster but takes more memory). Deprecated.",
                },
                "verbose": {
                    "type": "integer",
                    "default": 0,
                    "description": "Verbosity mode.",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Determines random number generation for centroid initialization",
                },
                "copy_x": {
                    "type": "boolean",
                    "default": True,
                    "description": """When pre-computing distances it is more numerically accurate to center the data first.
If copy_x is True (default), then the original data is not modified.
If False, the original data is modified, and put back before the function returns, but small numerical differences may be introduced by subtracting and then adding the data mean.
Note that if the original data is not C-contiguous, a copy will be made even if copy_x is False.
If the original data is sparse, but not in CSR format, a copy will be made even if copy_x is False.""",
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": 1,
                    "description": "The number of jobs to use for the computation. Deprecated.",
                },
                "algorithm": {
                    "description": """K-means algorithm to use.
The classical EM-style algorithm is “full”. The “elkan” variation is more efficient on data with well-defined clusters, by using the triangle inequality.
However it’s more memory intensive due to the allocation of an extra array of shape (n_samples, n_clusters).
For now “auto” (kept for backward compatibiliy) chooses “elkan” but it might change in the future for a better heuristic.""",
                    "enum": ["auto", "full", "elkan"],
                    "default": "auto",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Compute k-means clustering.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training instances to cluster. Array-like or sparse matrix, shape=(n_samples, n_features)",
        },
        "y": {
            "description": "not used, present here for API consistency by convention."
        },
        "sample_weight": {
            "anyOf": [{"type": "array", "items": {"type": "number"}}, {"enum": [None]}],
            "default": None,
            "description": "The weights for each observation in X",
        },
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Transform X to a cluster-distance space.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "New data to transform.",
        }
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "X transformed in the new space.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict the closest cluster each sample in X belongs to.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "New data to predict.",
        },
        "sample_weight": {
            "anyOf": [{"type": "array", "items": {"type": "number"}}, {"enum": [None]}],
            "default": None,
            "description": "The weights for each observation in X",
        },
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Index of the cluster each sample belongs to.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`KMeans`_ from scikit-learn.

.. _`KMeans`: https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.k_means.html",
    "import_from": "sklearn.cluster",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer", "clustering", "estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
KMeans = make_operator(SKLModel, _combined_schemas)

if sklearn.__version__ >= "1.0":
    # old: https://scikit-learn.org/0.24/modules/generated/sklearn.cluster.KMeans.html
    # new: https://scikit-learn.org/1.0/modules/generated/sklearn.cluster.KMeans.html
    KMeans = KMeans.customize_schema(
        precompute_distances=None,
        n_jobs=None,
        set_as_available=True,
    )

set_docstrings(KMeans)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sklearn
import sklearn.dummy

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "relevantToOptimizer": [],
            "additionalProperties": False,
            "required": ["strategy", "quantile"],
            "properties": {
                "strategy": {
                    "description": "Strategy to use to generate predictions.",
                    "anyOf": [
                        {
                            "enum": ["mean"],
                            "description": "Always predicts the mean of the training set.",
                        },
                        {
                            "enum": ["median"],
                            "description": "Always predicts the median of the training set.",
                        },
                        {
                            "enum": ["quantile"],
                            "description": "Always predicts a specified quantile of the training set, provided with the quantile parameter.",
                            "forOptimizer": False,
                        },
                        {
                            "enum": ["constant"],
                            "description": "Always predicts a constant label that is provided by the user. This is useful for metrics that evaluate a non-majority class",
                            "forOptimizer": False,
                        },
                    ],
                    "default": "mean",
                },
                "constant": {
                    "description": "The explicit constant as predicted by the “constant” strategy. This parameter is useful only for the “constant” strategy.",
                    "anyOf": [
                        {"type": ["integer", "string"]},
                        {"enum": [None]},
                        {"default": None},
                    ],
                },
                "quantile": {
                    "description": "The quantile to predict using the “quantile” strategy. A quantile of 0.5 corresponds to the median, while 0.0 to the minimum and 1.0 to the maximum.",
                    "anyOf": [
                        {"enum": [None]},
                        {"type": "number", "minimum": 0.0, "maximum": 1.0},
                    ],
                    "default": None,
                },
            },
        },
        {
            "description": "The constant strategy requires a non-None value for the constant hyperparameter.",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"strategy": {"not": {"enum": ["constant"]}}},
                },
                {
                    "type": "object",
                    "properties": {"constant": {"not": {"enum": [None]}}},
                },
            ],
        },
        {
            "description": "The quantile strategy requires a non-None value for the quantile hyperparameter.",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"strategy": {"not": {"enum": ["quantile"]}}},
                },
                {
                    "type": "object",
                    "properties": {"quantile": {"not": {"enum": [None]}}},
                },
            ],
        },
    ]
}

_input_fit_schema = {
    "required": ["X", "y"],
    "type": "object",
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array"},
        },
        "y": {
            "description": "Target values.",
            "type": "array",
            "items": {"type": "number"},
        },
    },
}

_input_predict_schema = {
    "type": "object",
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
        }
    },
}

_output_predict_schema = {
    "description": "Predicted values per sample.",
    "type": "array",
    "items": {"type": "number"},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Dummy regressor`_ regressor that makes predictions using simple rules.

.. _`Dummy regressor`: https://scikit-learn.org/stable/modules/generated/sklearn.dummy.DummyRegressor.html
""",
    "import_from": "sklearn.dummy",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}

DummyRegressor = lale.operators.make_operator(
    sklearn.dummy.DummyRegressor, _combined_schemas
)

lale.docstrings.set_docstrings(DummyRegressor)



from numpy import inf, nan
from sklearn.decomposition import TruncatedSVD as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _TruncatedSVDImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for TruncatedSVD    Dimensionality reduction using truncated SVD (aka LSA).",
    "allOf": [
        {
            "type": "object",
            "required": ["n_components", "algorithm", "n_iter", "random_state", "tol"],
            "relevantToOptimizer": ["n_components", "algorithm", "n_iter", "tol"],
            "additionalProperties": False,
            "properties": {
                "n_components": {
                    "type": "integer",
                    "minimun": 1,
                    "laleMaximum": "X/items/maxItems",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 256,
                    "distribution": "uniform",
                    "default": 2,
                    "description": "Desired dimensionality of output data",
                },
                "algorithm": {
                    "enum": ["arpack", "randomized"],
                    "default": "randomized",
                    "description": "SVD solver to use",
                },
                "n_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 5,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 5,
                    "description": "Number of iterations for randomized SVD solver",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.0,
                    "description": "Tolerance for ARPACK",
                },
            },
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit LSI model on training data X.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data.",
        },
        "y": {},
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Perform dimensionality reduction on X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "New data.",
        }
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Reduced version of X",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.TruncatedSVD#sklearn-decomposition-truncatedsvd",
    "import_from": "sklearn.decomposition",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}
TruncatedSVD = make_operator(_TruncatedSVDImpl, _combined_schemas)

set_docstrings(TruncatedSVD)



# Copyright 2019, 2020, 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aif360.algorithms.postprocessing

import lale.docstrings
import lale.operators

from .util import (
    _BasePostEstimatorImpl,
    _categorical_fairness_properties,
    _categorical_input_predict_schema,
    _categorical_output_predict_schema,
    _categorical_supervised_input_fit_schema,
)


class _EqOddsPostprocessingImpl(_BasePostEstimatorImpl):
    def __init__(
        self,
        *,
        favorable_labels,
        protected_attributes,
        unfavorable_labels=None,
        estimator,
        redact=True,
        seed=None,
    ):
        prot_attr_names = [pa["feature"] for pa in protected_attributes]
        unprivileged_groups = [{name: 0 for name in prot_attr_names}]
        privileged_groups = [{name: 1 for name in prot_attr_names}]
        mitigator = aif360.algorithms.postprocessing.EqOddsPostprocessing(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups,
            seed=seed,
        )
        super(_EqOddsPostprocessingImpl, self).__init__(
            favorable_labels=favorable_labels,
            protected_attributes=protected_attributes,
            unfavorable_labels=unfavorable_labels,
            estimator=estimator,
            redact=redact,
            mitigator=mitigator,
        )

    def predict_proba(self, X):
        raise NotImplementedError()


_input_fit_schema = _categorical_supervised_input_fit_schema
_input_predict_schema = _categorical_input_predict_schema
_output_predict_schema = _categorical_output_predict_schema

_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                *_categorical_fairness_properties.keys(),
                "estimator",
                "redact",
                "seed",
            ],
            "relevantToOptimizer": [],
            "properties": {
                **_categorical_fairness_properties,
                "estimator": {
                    "description": "Nested supervised learning operator for which to mitigate fairness.",
                    "laleType": "operator",
                },
                "redact": {
                    "description": "Whether to redact protected attributes before data preparation (recommended) or not.",
                    "type": "boolean",
                    "default": True,
                },
                "seed": {
                    "description": "Seed to make `predict` repeatable.",
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": None,
                },
            },
        }
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Equalized odds postprocessing`_ post-estimator fairness mitigator. Solves a linear program to find probabilities with which to change output labels to optimize equalized odds (`Hardt et al. 2016`_, `Pleiss et al. 2017`_).

.. _`Equalized odds postprocessing`: https://aif360.readthedocs.io/en/latest/modules/generated/aif360.algorithms.postprocessing.EqOddsPostprocessing.html
.. _`Hardt et al. 2016`: https://papers.nips.cc/paper/2016/hash/9d2682367c3935defcb1f9e247a97c0d-Abstract.html
.. _`Pleiss et al. 2017`: https://proceedings.neurips.cc/paper/2017/hash/b8b9c74ac526fffbeb2d39ab038d1cd7-Abstract.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.aif360.eq_odds_postprocessing.html#lale.lib.aif360.eq_odds_postprocessing.EqOddsPostprocessing",
    "import_from": "aif360.algorithms.postprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier", "interpretable"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}

EqOddsPostprocessing = lale.operators.make_operator(
    _EqOddsPostprocessingImpl, _combined_schemas
)

lale.docstrings.set_docstrings(EqOddsPostprocessing)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import lale.docstrings
import lale.operators

try:
    import lightgbm
    import lightgbm.sklearn

    lightgbm_installed = True
except ImportError:
    lightgbm_installed = False
    if TYPE_CHECKING:
        import lightgbm  # type: ignore


class _LGBMRegressorImpl:
    def __init__(
        self,
        boosting_type="gbdt",
        num_leaves=31,
        max_depth=-1,
        learning_rate=0.1,
        n_estimators=100,
        subsample_for_bin=200000,
        objective=None,
        class_weight=None,
        min_split_gain=0.0,
        min_child_weight=0.001,
        min_child_samples=20,
        subsample=1.0,
        subsample_freq=0,
        colsample_bytree=1.0,
        reg_alpha=0.0,
        reg_lambda=0.0,
        random_state=None,
        n_jobs=-1,
        silent=True,
        importance_type="split",
    ):
        assert lightgbm_installed, """Your Python environment does not have lightgbm installed. You can install it with
    pip install lightgbm
or with
    pip install 'lale[full]'"""
        self._hyperparams = {
            "boosting_type": boosting_type,
            "num_leaves": num_leaves,
            "max_depth": max_depth,
            "learning_rate": learning_rate,
            "n_estimators": n_estimators,
            "subsample_for_bin": subsample_for_bin,
            "objective": objective,
            "class_weight": class_weight,
            "min_split_gain": min_split_gain,
            "min_child_weight": min_child_weight,
            "min_child_samples": min_child_samples,
            "subsample": subsample,
            "subsample_freq": subsample_freq,
            "colsample_bytree": colsample_bytree,
            "reg_alpha": reg_alpha,
            "reg_lambda": reg_lambda,
            "random_state": random_state,
            "n_jobs": n_jobs,
            "silent": silent,
            "importance_type": importance_type,
        }
        self._wrapped_model = lightgbm.sklearn.LGBMRegressor(**self._hyperparams)

    def fit(self, X, y=None, **fit_params):
        try:
            self._wrapped_model.fit(X, y, **fit_params)
        except Exception as e:
            raise RuntimeError(str(self._hyperparams)) from e
        return self

    def predict(self, X, **predict_params):
        return self._wrapped_model.predict(X, **predict_params)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

    def score(self, X, y):
        from sklearn.metrics import r2_score

        y_pred = self.predict(X)
        return r2_score(y, y_pred)


_hyperparams_schema = {
    "description": "LightGBM classifier. (https://lightgbm.readthedocs.io/en/latest/Python-API.html#scikit-learn-api)",
    "allOf": [
        {
            "type": "object",
            "relevantToOptimizer": [
                "boosting_type",
                "num_leaves",
                "learning_rate",
                "n_estimators",
                "min_child_weight",
                "min_child_samples",
                "subsample",
                "subsample_freq",
                "colsample_bytree",
                "reg_alpha",
                "reg_lambda",
            ],
            "additionalProperties": False,
            "properties": {
                "boosting_type": {
                    "anyOf": [
                        {
                            "enum": ["gbdt"],
                            "description": "Traditional Gradient Boosting Decision Tree.",
                        },
                        {
                            "enum": ["dart"],
                            "description": "Dropouts meet Multiple Additive Regression Trees.",
                        },
                        {
                            "enum": ["goss"],
                            "forOptimizer": False,
                            "description": "Gradient-based One-Side Sampling.",
                        },
                        {
                            "enum": ["rf"],
                            "forOptimizer": False,
                            "description": "Random Forest.",
                        },
                    ],
                    "default": "gbdt",
                },
                "num_leaves": {
                    "anyOf": [
                        {"type": "integer", "forOptimizer": False},
                        {"enum": [2, 4, 8, 32, 64, 128, 16]},
                    ],
                    "default": 31,
                    "description": "Maximum tree leaves for base learners",
                },
                "max_depth": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 5,
                        }
                    ],
                    "default": -1,
                    "description": "Maximum tree depth for base learners, <=0 means no limit",
                },
                "learning_rate": {
                    "type": "number",
                    "minimumForOptimizer": 0.02,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.1,
                    "description": "Boosting learning rate.",
                },
                "n_estimators": {
                    "type": "integer",
                    "minimumForOptimizer": 50,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 200,
                    "description": "Number of boosted trees to fit.",
                },
                "subsample_for_bin": {
                    "type": "integer",
                    "default": 200000,
                    "description": "Number of samples for constructing bins.",
                },
                "objective": {
                    "anyOf": [{"type": "object"}, {"enum": ["regression", None]}],
                    "default": None,
                    "description": "Specify the learning task and the corresponding learning objective or a custom objective function to be used",
                },
                "class_weight": {  # Should not apply to regression, but documentation includes it.
                    "anyOf": [{"type": "object"}, {"enum": ["balanced", None]}],
                    "default": None,
                    "description": "Weights associated with classes",
                },
                "min_split_gain": {
                    "type": "number",
                    "default": 0.0,
                    "description": "Minimum loss reduction required to make a further partition on a leaf node of the tree.",
                },
                "min_child_weight": {
                    "type": "number",
                    "minimumForOptimizer": 0.0001,
                    "maximumForOptimizer": 0.01,
                    "default": 1e-3,
                    "description": "Minimum sum of instance weight (hessian) needed in a child (leaf).",
                },
                "min_child_samples": {
                    "type": "integer",
                    "minimumForOptimizer": 5,
                    "maximumForOptimizer": 30,
                    "distribution": "uniform",
                    "default": 20,
                    "description": "Minimum number of data needed in a child (leaf).",
                },
                "subsample": {
                    "type": "number",
                    "minimumForOptimizer": 0.01,
                    "maximumForOptimizer": 1.0,
                    "distribution": "uniform",
                    "default": 1.0,
                    "description": "Subsample ratio of the training instance.",
                },
                "subsample_freq": {
                    "type": "integer",
                    "minimumForOptimizer": 0,
                    "maximumForOptimizer": 5,
                    "distribution": "uniform",
                    "default": 0,
                    "description": "Frequence of subsample, <=0 means no enable.",
                },
                "colsample_bytree": {
                    "type": "number",
                    "default": 1.0,
                    "minimumForOptimizer": 0.01,
                    "maximumForOptimizer": 1.0,
                    "description": "Subsample ratio of columns when constructing each tree.",
                },
                "reg_alpha": {
                    "type": "number",
                    "minimumForOptimizer": 0.0,
                    "maximumForOptimizer": 1.0,
                    "default": 0.0,
                    "description": "L1 regularization term on weights.",
                },
                "reg_lambda": {
                    "type": "number",
                    "minimumForOptimizer": 0.0,
                    "maximumForOptimizer": 1.0,
                    "default": 0.0,
                    "description": "L2 regularization term on weights.",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Random number seed. If None, default seeds in C++ code will be used.",
                },
                "n_jobs": {
                    "type": "integer",
                    "default": -1,
                    "description": "Number of parallel threads.",
                },
                "silent": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to print messages while running boosting.",
                },
                "importance_type": {
                    "enum": ["split", "gain"],
                    "default": "split",
                    "description": "The type of feature importance to be filled into `feature_importances_`.",
                },
            },
        },
        {
            "description": "boosting_type `rf` needs bagging (which means subsample_freq > 0 and subsample < 1.0)",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"boosting_type": {"not": {"enum": ["rf"]}}},
                },
                {
                    "allOf": [
                        {
                            "type": "object",
                            "properties": {"subsample_freq": {"not": {"enum": [0]}}},
                        },
                        {
                            "type": "object",
                            "properties": {"subsample": {"not": {"enum": [1.0]}}},
                        },
                    ]
                },
            ],
        },
        {
            "description": "boosting_type `goss` cannot use bagging (which means subsample_freq = 0 and subsample = 1.0)",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"boosting_type": {"not": {"enum": ["goss"]}}},
                },
                {
                    "type": "object",
                    "properties": {"subsample_freq": {"enum": [0]}},
                },
                {
                    "type": "object",
                    "properties": {"subsample": {"enum": [1.0]}},
                },
            ],
        },
    ],
}

_input_fit_schema = {
    "description": "Build a lightgbm model from the training set (X, y).",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The input samples. Internally, it will be converted to",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values real numbers",
        },
        "sample_weight": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "number"},
                },
                {"enum": [None]},
            ],
            "default": None,
            "description": "Weights of training data.",
        },
        "init_score": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "number"},
                },
                {"enum": [None]},
            ],
            "default": None,
            "description": "Init score of training data.",
        },
        "group": {"default": None, "description": "Group data of training data."},
        "eval_set": {
            "default": None,
            "description": "A list of (X, y) tuple pairs to use as validation sets.",
        },
        "eval_names": {"default": None, "description": "Names of eval_set."},
        "eval_sample_weight": {"default": None, "description": "Weights of eval data."},
        "eval_class_weight": {
            "anyOf": [{"type": "array", "items": {"type": "number"}}, {"enum": [None]}],
            "default": None,
            "description": "Class weights of eval data.",
        },
        "eval_init_score": {"default": None, "description": "Init score of eval data."},
        "eval_group": {"default": None, "description": "Group data of eval data."},
        "eval_metric": {
            "anyOf": [
                {"type": "array", "items": {"type": "string"}},
                {"enum": ["l2", None]},
                {"laleType": "callable"},
            ],
            "default": None,
            "description": "string, list of strings, callable or None, optional (default=None).",
        },
        "early_stopping_rounds": {
            "anyOf": [{"type": "integer"}, {"enum": [None]}],
            "default": None,
            "description": "Activates early stopping. The model will train until the validation score stops improving.",
        },
        "verbose": {
            "anyOf": [{"type": "boolean"}, {"type": "integer"}],
            "default": True,
            "description": "Requires at least one evaluation data.",
        },
        "feature_name": {
            "anyOf": [
                {"type": "array", "items": {"type": "string"}},
                {"enum": ["auto"]},
            ],
            "default": "auto",
            "description": "Feature names. If ‘auto’ and data is pandas DataFrame, data columns names are used.",
        },
        "categorical_feature": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"anyOf": [{"type": "string"}, {"type": "integer"}]},
                },
                {"enum": ["auto"]},
            ],
            "default": "auto",
            "description": "Categorical features. If list of int, interpreted as indices. If list of strings, interpreted as feature names.",
        },
        "callbacks": {
            "anyOf": [{"type": "array", "items": {"type": "object"}}, {"enum": [None]}],
            "default": None,
            "description": "List of callback functions that are applied at each iteration. ",
        },
    },
}
_input_predict_schema = {
    "description": "Return the predicted value for each sample.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": " Input features matrix.",
        },
        "raw_score": {
            "type": "boolean",
            "default": False,
            "description": "Whether to predict raw scores.",
        },
        "num_iteration": {
            "anyOf": [{"type": "integer"}, {"enum": [None]}],
            "default": None,
            "description": "Limit number of iterations in the prediction.",
        },
        "pred_leaf": {
            "type": "boolean",
            "default": False,
            "description": "Whether to predict leaf index.",
        },
        "pred_contrib": {
            "type": "boolean",
            "default": False,
            "description": "Whether to predict feature contributions.",
        },
    },
}
_output_predict_schema = {
    "description": "Return the predicted value for each sample.",
    "type": "array",
    "items": {"type": "number"},
}
_input_predict_proba_schema = {
    "description": "Return the predicted probability for each class for each sample.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": " Input features matrix.",
        },
        "raw_score": {
            "type": "boolean",
            "default": False,
            "description": "Whether to predict raw scores.",
        },
        "num_iteration": {
            "anyOf": [{"type": "integer"}, {"enum": [None]}],
            "default": None,
            "description": "Limit number of iterations in the prediction.",
        },
        "pred_leaf": {
            "type": "boolean",
            "default": False,
            "description": "Whether to predict leaf index.",
        },
        "pred_contrib": {
            "type": "boolean",
            "default": False,
            "description": "Whether to predict feature contributions.",
        },
    },
}
_output_predict_proba_schema = {
    "description": "Return the predicted probability for each class for each sample.",
    "type": "array",
    "items": {
        "type": "array",
        "items": {"type": "number"},
    },
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.lightgbm.lgbm_regressor.html",
    "import_from": "lightgbm.sklearn",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}


LGBMRegressor = lale.operators.make_operator(_LGBMRegressorImpl, _combined_schemas)

if lightgbm_installed and lightgbm.__version__ >= "3.3.0":  #
    # https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html
    LGBMRegressor = LGBMRegressor.customize_schema(
        silent={
            "description": "Whether to print messages while running boosting.",
            "anyOf": [{"enum": ["warn"]}, {"type": "boolean"}],
            "default": "warn",
        },
        set_as_available=True,
    )

lale.docstrings.set_docstrings(LGBMRegressor)



from numpy import inf, nan
from sklearn.neighbors import NearestCentroid as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _NearestCentroidImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for NearestCentroid    Nearest centroid classifier.",
    "allOf": [
        {
            "type": "object",
            "required": ["metric", "shrink_threshold"],
            "relevantToOptimizer": ["metric", "shrink_threshold"],
            "additionalProperties": False,
            "properties": {
                "metric": {
                    "anyOf": [
                        {"laleType": "callable", "forOptimizer": False},
                        {
                            "enum": [
                                "cityblock",
                                "cosine",
                                "euclidean",
                                "l1",
                                "l2",
                                "manhattan",
                                "braycurtis",
                                "canberra",
                                "chebyshev",
                                "correlation",
                                "dice",
                                "hamming",
                                "jaccard",
                                "kulsinski",
                                "mahalanobis",
                                "minkowski",
                                "rogerstanimoto",
                                "russellrao",
                                "seuclidean",
                                "sokalmichener",
                                "sokalsneath",
                                "sqeuclidean",
                                "yule",
                            ]
                        },
                    ],
                    "default": "euclidean",
                    "description": "The metric to use when calculating distance between instances in a feature array",
                },
                "shrink_threshold": {
                    "anyOf": [
                        {
                            "type": "number",
                            "minimumForOptimizer": 0.0,
                            "maximumForOptimizer": 1.0,
                            "distribution": "uniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Threshold for shrinking centroids to remove features.",
                },
            },
        },
        {
            "description": "threshold shrinking not supported for sparse input",
            "anyOf": [
                {"type": "object", "laleNot": "X/isSparse"},
                {
                    "type": "object",
                    "properties": {"shrink_threshold": {"enum": [None]}},
                },
            ],
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the NearestCentroid model according to the given training data.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training vector, where n_samples is the number of samples and n_features is the number of features",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values (integers)",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Perform classification on an array of test vectors X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}}
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Perform classification on an array of test vectors X.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.neighbors.NearestCentroid#sklearn-neighbors-nearestcentroid",
    "import_from": "sklearn.neighbors",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
NearestCentroid = make_operator(_NearestCentroidImpl, _combined_schemas)

set_docstrings(NearestCentroid)



from numpy import inf, nan
from sklearn.linear_model import SGDRegressor as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _SGDRegressorImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for SGDRegressor    Linear model fitted by minimizing a regularized empirical loss with SGD",
    "allOf": [
        {
            "type": "object",
            "required": [
                "loss",
                "penalty",
                "alpha",
                "l1_ratio",
                "fit_intercept",
                "max_iter",
                "tol",
                "shuffle",
                "verbose",
                "epsilon",
                "random_state",
                "learning_rate",
                "eta0",
                "power_t",
                "early_stopping",
                "validation_fraction",
                "n_iter_no_change",
                "warm_start",
                "average",
            ],
            "relevantToOptimizer": [
                "loss",
                "penalty",
                "alpha",
                "fit_intercept",
                "max_iter",
                "tol",
                "shuffle",
                "epsilon",
                "learning_rate",
                "eta0",
            ],
            "additionalProperties": False,
            "properties": {
                "loss": {
                    "enum": [
                        "epsilon_insensitive",
                        "huber",
                        "squared_epsilon_insensitive",
                        "squared_loss",
                    ],
                    "default": "squared_loss",
                    "description": "The loss function to be used",
                },
                "penalty": {
                    "XXX TODO XXX": "str, 'none', 'l2', 'l1', or 'elasticnet'",
                    "description": "The penalty (aka regularization term) to be used",
                    "enum": ["elasticnet", "l1", "l2", "none"],
                    "default": "l2",
                },
                "alpha": {
                    "type": "number",
                    "minimumForOptimizer": 1e-10,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "Constant that multiplies the regularization term",
                },
                "l1_ratio": {
                    "type": "number",
                    "default": 0.15,
                    "description": "The Elastic Net mixing parameter, with 0 <= l1_ratio <= 1",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether the intercept should be estimated or not",
                },
                "max_iter": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 10,
                            "maximumForOptimizer": 1000,
                            "distribution": "uniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The maximum number of passes over the training data (aka epochs)",
                },
                "tol": {
                    "anyOf": [
                        {
                            "type": "number",
                            "minimumForOptimizer": 1e-08,
                            "maximumForOptimizer": 0.01,
                            "distribution": "loguniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The stopping criterion",
                },
                "shuffle": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether or not the training data should be shuffled after each epoch",
                },
                "verbose": {
                    "type": "integer",
                    "default": 0,
                    "description": "The verbosity level.",
                },
                "epsilon": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 1.35,
                    "distribution": "loguniform",
                    "default": 0.1,
                    "description": "Epsilon in the epsilon-insensitive loss functions; only if `loss` is 'huber', 'epsilon_insensitive', or 'squared_epsilon_insensitive'",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The seed of the pseudo random number generator to use when shuffling the data",
                },
                "learning_rate": {
                    "enum": ["adaptive", "constant", "optimal", "invscaling"],
                    "default": "invscaling",
                    "description": "The learning rate schedule:  'constant':     eta = eta0 'optimal':     eta = 1.0 / (alpha * (t + t0))     where t0 is chosen by a heuristic proposed by Leon Bottou",
                },
                "eta0": {
                    "type": "number",
                    "minimumForOptimizer": 0.01,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.01,
                    "description": "The initial learning rate for the 'constant', 'invscaling' or 'adaptive' schedules",
                },
                "power_t": {
                    "type": "number",
                    "default": 0.25,
                    "description": "The exponent for inverse scaling learning rate [default 0.5].",
                },
                "early_stopping": {
                    "type": "boolean",
                    "default": False,
                    "description": "Whether to use early stopping to terminate training when validation score is not improving",
                },
                "validation_fraction": {
                    "type": "number",
                    "default": 0.1,
                    "description": "The proportion of training data to set aside as validation set for early stopping",
                },
                "n_iter_no_change": {
                    "type": "integer",
                    "default": 5,
                    "description": "Number of iterations with no improvement to wait before early stopping",
                },
                "warm_start": {
                    "type": "boolean",
                    "default": False,
                    "description": "When set to True, reuse the solution of the previous call to fit as initialization, otherwise, just erase the previous solution",
                },
                "average": {
                    "anyOf": [{"type": "boolean"}, {"type": "integer"}],
                    "default": False,
                    "description": "When set to True, computes the averaged SGD weights and stores the result in the ``coef_`` attribute",
                },
            },
        },
        {
            "XXX TODO XXX": "Parameter: max_iter > only impacts the behavior in the fit method, and not the partial_fit"
        },
        {
            "description": "epsilon, only if loss is 'huber', 'epsilon_insensitive', or 'squared_epsilon_insensitive'",
            "anyOf": [
                {"type": "object", "properties": {"epsilon": {"enum": [0.1]}}},
                {
                    "type": "object",
                    "properties": {
                        "loss": {
                            "enum": [
                                "huber",
                                "epsilon_insensitive",
                                "squared_epsilon_insensitive",
                            ]
                        }
                    },
                },
            ],
        },
        {
            "description": "validation_fraction, only used if early_stopping is true",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"validation_fraction": {"enum": [0.1]}},
                },
                {"type": "object", "properties": {"early_stopping": {"enum": [True]}}},
            ],
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit linear model with Stochastic Gradient Descent.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values",
        },
        "coef_init": {
            "type": "array",
            "items": {"type": "number"},
            "description": "The initial coefficients to warm-start the optimization.",
        },
        "intercept_init": {
            "type": "array",
            "items": {"type": "number"},
            "description": "The initial intercept to warm-start the optimization.",
        },
        "sample_weight": {
            "anyOf": [{"type": "array", "items": {"type": "number"}}, {"enum": [None]}],
            "default": None,
            "description": "Weights applied to individual samples (1",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the linear model",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}}
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predicted target values per element in X.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.SGDRegressor#sklearn-linear_model-sgdregressor",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
SGDRegressor = make_operator(_SGDRegressorImpl, _combined_schemas)

set_docstrings(SGDRegressor)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.preprocessing

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Scale features using statistics that are robust to outliers.",
    "allOf": [
        {
            "type": "object",
            "required": ["quantile_range", "copy"],
            "relevantToOptimizer": ["with_centering", "with_scaling", "quantile_range"],
            "additionalProperties": False,
            "properties": {
                "with_centering": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, center the data before scaling.",
                },
                "with_scaling": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, scale the data to interquartile range.",
                },
                "quantile_range": {
                    "type": "array",
                    "laleType": "tuple",
                    "minItemsForOptimizer": 2,
                    "maxItemsForOptimizer": 2,
                    "items": [
                        {
                            "type": "number",
                            "minimumForOptimizer": 0.001,
                            "maximumForOptimizer": 0.3,
                        },
                        {
                            "type": "number",
                            "minimumForOptimizer": 0.7,
                            "maximumForOptimizer": 0.999,
                        },
                    ],
                    "default": [0.25, 0.75],
                    "description": "Default: (25.0, 75.0) = (1st quantile, 3rd quantile) = IQR",
                },
                "copy": {
                    "type": "boolean",
                    "default": True,
                    "description": "If False, try to avoid a copy and do inplace scaling instead.",
                },
            },
        },
        {
            "description": "Cannot center sparse matrices: use `with_centering=False` instead. See docstring for motivation and alternatives.",
            "anyOf": [
                {"type": "object", "properties": {"with_centering": {"enum": [False]}}},
                {"type": "object", "laleNot": "X/isSparse"},
            ],
        },
    ],
}
_input_fit_schema = {
    "description": "Compute the median and quantiles to be used for scaling.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The data used to compute the median and quantiles",
        },
        "y": {},
    },
}
_input_transform_schema = {
    "description": "Center and scale the data.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The data used to scale along the specified axis.",
        },
    },
}
_output_transform_schema = {
    "description": "Center and scale the data.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Robust scaler`_ transformer from scikit-learn.

.. _`Robust scaler`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.robust_scaler.html",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

RobustScaler: lale.operators.PlannedIndividualOp
RobustScaler = lale.operators.make_operator(
    sklearn.preprocessing.RobustScaler, _combined_schemas
)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.RobustScaler.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.preprocessing.RobustScaler.html
    from lale.schemas import Bool

    RobustScaler = RobustScaler.customize_schema(
        unit_variance=Bool(
            desc="If True, scale data so that normally distributed features have a variance of 1. In general, if the difference between the x-values of q_max and q_min for a standard normal distribution is greater than 1, the dataset will be scaled down. If less than 1, the dataset will be scaled up.",
            default=False,
            forOptimizer=True,
        ),
        set_as_available=True,
    )


lale.docstrings.set_docstrings(RobustScaler)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

# import sys
from datetime import datetime

from setuptools import find_packages, setup

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())

try:
    import builtins

    # This trick is borrowed from scikit-learn
    # This is a bit (!) hackish: we are setting a global variable so that the
    # main lale __init__ can detect if it is being loaded by the setup
    # routine, to avoid attempting to import components before installation.
    builtins.__LALE_SETUP__ = True  # type: ignore
except ImportError:
    pass

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

on_rtd = os.environ.get("READTHEDOCS") == "True"
if on_rtd:
    install_requires = []
else:
    install_requires = [
        "numpy",
        "black==19.10b0",
        "graphviz",
        "hyperopt<=0.2.5",
        "jsonschema",
        "jsonsubschema",
        "scikit-learn",
        "scipy",
        "pandas",
        "decorator",
        "astunparse",
    ]

import lale  # noqa: E402

if "TRAVIS" in os.environ:
    now = datetime.now().strftime("%y%m%d%H%M")
    VERSION = f"{lale.__version__}-{now}"
else:
    VERSION = lale.__version__

extras_require = {
    "full": [
        "xgboost<=1.5.1",
        "lightgbm",
        "snapml>=1.7.0rc3<=1.8.0",
        "liac-arff>=2.4.0",
        "tensorflow>=2.4.0",
        "smac<=0.10.0",
        "numba",
        "aif360>=0.4.0",
        "torch>=1.0",
        "BlackBoxAuditing",
        "imbalanced-learn",
        "cvxpy>=1.0",
        "fairlearn",
        "h5py",
    ],
    "dev": ["pre-commit"],
    "test": [
        # "autoai-libs>=1.12.6",
        "joblib",
        "jupyter",
        "numpydoc",
        "sphinx==2.4.4",
        "m2r",
        "sphinx_rtd_theme",
        "sphinxcontrib.apidoc",
        "pytest",
        "pyspark",
        "func_timeout",
    ],
    "fairness": [
        "liac-arff>=2.4.0",
        "aif360<0.4.0",
        "BlackBoxAuditing",
    ],
}

# # Inserting tensorflow 1.x only for Python 3.7
# if sys.version_info.minor == 7:
#     extras_require["full"].append("tensorflow>=1.13.1,<2")

setup(
    name="lale",
    version=VERSION,
    author="Guillaume Baudart, Martin Hirzel, Kiran Kate, Parikshit Ram, Avraham Shinnar",
    description="Library for Semi-Automated Data Science",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/lale",
    python_requires=">=3.6",
    packages=find_packages(),
    license="",
    install_requires=install_requires,
    extras_require=extras_require,
)
